import gzip
import os
import six
import warnings
import pandas as pd
from io import TextIOWrapper

try:
    import pyarrow.parquet as pq
    parquet = True
except ImportError:
    parquet = False

try:
    import fastavro as avro
    avro_imported = True
except ImportError:
    avro_imported = False


def read_df(datafile, **read_args):
    """ Read a dataframe from file based on the file extension

    The following formats are supported:
    parquet, avro, csv, pickle

    Parameters
    ----------
    datafile : DataFile
        DataFile instance with the path and handle
    read_args : optional
        All keyword args are passed to the read function

    Returns
    -------
    data : pd.DataFrame

    Notes
    -----
    The read functions are taken from pandas, e.g. pd.read_csv
    Check the pandas doc for more information on the supported arguments

    """
    datafile.handle.seek(0)  # ensure we start from the beginning of the file
    filetype, compression = _get_extension(datafile.path)
    reader = _readers[filetype]
    if reader in (pd.read_csv, pd.read_pickle) and compression is not None:
        read_args["compression"] = compression
    return reader(datafile.handle, **read_args)


def write_df(df, datafile, **write_args):
    """ Write a dataframe to file based on the file extension

    The following formats are supported:
    parquet, avro, csv, pickle

    Parameters
    ----------
    df : pd.DataFrame
        The dataframe to write to disk
    datafile : DataFile
        Datafile instance with the path and file handle
    write_args : optional
        All keyword args are passed to the write function

    Notes
    -----
    The write functions are taken from pandas, e.g. pd.to_csv
    Check the pandas doc for more information on the supported arguments

    """
    extension, compression = _get_extension(datafile.path)
    write_name = _writers[extension]
    # infer compression from filepath or from explicit arg
    compression = compression or write_args.get('compression')
    buffer = datafile.handle

    # Some customizations for different file types
    if write_name == 'to_avro':
        return _write_avro(df, buffer, **write_args)

    if write_name == 'to_parquet' and not pd.Series(df.columns).map(type).eq(str).all():
        warnings.warn(
            'Dataframe contains non-string column names, which cannot be saved in parquet.\n'
            'Blocks will attempt to convert them to strings.'
        )
        df.columns = df.columns.astype('str')

    if write_name == 'to_csv':
        # make index=False the default for similar behaviour to other formats
        write_args['index'] = write_args.get('index', False)

    # For csv and pickle we have to manually compress
    if compression == 'gzip' and write_name in ('to_csv', 'to_pickle'):
        buffer = gzip.GzipFile(fileobj=buffer, mode='w')
    elif compression is not None and write_name in ('to_csv', 'to_pickle'):
        raise ValueError('Compression {} is not supported for CSV/Pickle'.format(compression))

    # And for 23 compatibility need a textiowrapper for csv
    if write_name == 'to_csv' and six.PY3:
        buffer = TextIOWrapper(buffer)

    write_fn = getattr(df, write_name)
    return write_fn(buffer, **write_args)


def _read_parquet(path, **read_args):
    """ Read a dataframe from parquet file

    Parameters
    ----------
    path : str
        Path to the parquet file
    read_args : optional
        All keyword args are passed to pyarrow.parquet.read_table

    Returns
    -------
    data : pd.DataFrame

    """
    if not parquet:
        raise ImportError('Parquet support requires pyarrow.\n'
                          'Install blocks with the [pq] option or `pip install pyarrow`')
    return pq.read_table(path, **read_args).to_pandas()


def _read_avro(handle, **read_args):
    if not avro_imported:
        raise ImportError('Avro support requires fastavro.\n'
                          'Install blocks with the [avro] option or `pip install fastavro`')
    records = []
    avro_reader = avro.reader(handle)
    for record in avro_reader:
        records.append(record)
    return pd.DataFrame.from_dict(records)


def _write_avro(df, handle, **write_args):
    if not avro_imported:
        raise ImportError('Avro support requires fastavro.\n'
                          'Install blocks with the [avro] option or `pip install fastavro`')
    schema = None
    schema_path = None
    try:
        schema = write_args['schema']
    except KeyError:
        try:
            schema_path = write_args['schema_path']
        except KeyError:
            raise Exception("You must provide a schema or schema path when writing to Avro")
    if schema is None:
        schema = avro.schema.load_schema(schema_path)
    records = df.to_dict('records')
    avro.writer(handle, schema, records)


def _get_extension(path):
    name, ext = os.path.splitext(path)
    # Support compression extensions, eg part.csv.gz
    comp = None
    if ext in _compressions and '.' in name:
        comp = ext
        name, ext = os.path.splitext(name)
    return ext, _compressions.get(comp)


_readers = {
    '.pq': _read_parquet,
    '.parquet': _read_parquet,
    '.csv': pd.read_csv,
    '.pkl': pd.read_pickle,
    '.avro': _read_avro,
}


_writers = {
    '.pq': 'to_parquet',
    '.parquet': 'to_parquet',
    '.csv': 'to_csv',
    '.pkl': 'to_pickle',
    '.avro': 'to_avro',
}

_compressions = {
    '.gz': 'gzip',
    '.bz2': 'bz2',
    '.zip': 'zip',
    '.xz': 'xz'
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from blocks import core


def test_iterate(populated):
    dfs = core.iterate(populated)
    for i in range(4):
        for j in range(4):
            cname, rname, df = next(dfs)
            assert(cname == 'c{}'.format(i))
            assert(rname == 'part.{}.csv'.format(j))
            assert(df.shape == (10, 11))


def test_iterate_ordered(populated):
    order = ['c3', 'c1', 'c2', 'c0']
    dfs = core.iterate(populated, cgroups=order)
    for i in range(4):
        for j in range(4):
            cname, rname, df = next(dfs)
            assert(cname == order[i])
            assert(rname == 'part.{}.csv'.format(j))
            assert(df.shape == (10, 11))


def test_iterate_axis0(populated):
    dfs = core.iterate(populated, axis=0)
    for i in range(4):
        rname, df = next(dfs)
        assert(rname == 'part.{}.csv'.format(i))
        assert(df.shape == (10, 41))


def test_iterate_axis1(populated):
    dfs = core.iterate(populated, axis=1)
    for i in range(4):
        cname, df = next(dfs)
        assert(cname == 'c{}'.format(i))
        assert(df.shape == (40, 11))



import os
import pytest
import pandas as pd
import numpy as np

from blocks.filesystem import GCSFileSystem, GCSNativeFileSystem
from delegator import run

BUCKET = 'gs://blocks-example'

if os.environ.get('CI'):
    inputs = ['local']
    outputs = ['local']
    filesystems = ['gcs']
else:
    inputs = ['local', 'gcs', 'gcs_extra']
    outputs = ['local', 'gcs']
    filesystems = ['gcs', 'native']


@pytest.fixture(scope='session', params=filesystems)
def fs(request):
    if request.param == 'gcs':
        return GCSFileSystem()

    if request.param == 'native':
        return GCSNativeFileSystem()


@pytest.fixture(scope='function')
def gcstemp():
    path = os.path.join(BUCKET, 'temp')
    yield path
    run('gsutil -m rm -r {}'.format(path))


@pytest.fixture(scope='session')
def populated_local(request, tmpdir_factory):
    tmpdir = str(tmpdir_factory.mktemp('data'))
    _populate(tmpdir)
    return tmpdir


# This is the same directory structure as above but paramatrized on different file systems
@pytest.fixture(scope='session', params=inputs)
def populated(request, populated_local):
    if request.param == 'local':
        yield populated_local

    if request.param == 'gcs':
        path = os.path.join(BUCKET, 'data1')
        run('gsutil -m cp -r {} {}'.format(populated_local, path))
        yield path
        run('gsutil -m rm -r {}'.format(path))

    if request.param == 'gcs_extra':
        path = os.path.join(BUCKET, 'data2')
        run('gsutil -m cp -r {} {}'.format(populated_local, path))
        # Also add an extra file
        run('touch extra')
        run('gsutil cp extra {}'.format(path))
        os.remove('extra')
        yield path
        run('gsutil -m rm -r {}'.format(path))


@pytest.fixture(scope='session')
def keys():
    return pd.Series(['key{:02d}'.format(i) for i in range(40)])


@pytest.fixture()
def randomdata():
    df = pd.DataFrame(
        np.random.rand(10, 10),
        columns=['f{}'.format(i) for i in range(10)]
    )
    return df


@pytest.fixture()
def datadir_local(request, tmpdir_factory):
    return str(tmpdir_factory.mktemp('data'))


@pytest.fixture(params=outputs)
def datadir(request, tmpdir_factory):
    if request.param == 'local':
        tmpdir = str(tmpdir_factory.mktemp('data'))
        yield tmpdir

    if request.param == 'gcs':
        path = os.path.join(BUCKET, 'output')
        yield path
        run('gsutil -m rm -r {}'.format(path))


def _populate(tmpdir):
    """ Create a directory of blocks with 4 cgroups and 4 rgroups
    """
    for c in range(4):
        cgroup = os.path.join(tmpdir, 'c{}'.format(c))
        if not os.path.exists(cgroup):
            os.makedirs(cgroup)
        for r in range(4):
            df = pd.DataFrame(
                np.random.rand(10, 10),
                index=list(range(r*10, (r+1)*10)),
                columns=['f{}_{}'.format(c, i) for i in range(10)]
            )
            df['key'] = ['key{:02d}'.format(i) for i in df.index]  # common key for merges
            df.to_csv(os.path.join(cgroup, 'part.{}.csv'.format(r)), index=False)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import pytest


TEST_STRING = b'test'


def test_ls_directory(populated, fs):
    ex = os.path.join(populated, '')
    expected = [os.path.join(populated, 'c{}'.format(i)) for i in range(4)]

    # one of the fixtures has an extra file
    found = fs.ls(ex)
    if len(found) > 4:
        found = found[:-1]
    assert(found == expected)


def test_ls_wildcard(populated, fs):
    ex = os.path.join(populated, '*/part.1.csv')
    expected = [os.path.join(populated, 'c{}/part.1.csv'.format(i)) for i in range(4)]
    assert(fs.ls(ex) == expected)


def test_ls_double_wildcard(populated, fs):
    ex = os.path.join(populated, '**')
    expected = {os.path.join(populated, 'c{}/part.{}.csv'.format(i, j))
                for i in range(4)
                for j in range(4)}
    assert(expected.issubset(set(fs.ls(ex))))


def test_ls_pattern(populated, fs):
    ex = os.path.join(populated, '*/part.[01].csv')
    expected = [os.path.join(populated, 'c{}/part.{}.csv'.format(i, j))
                for i in range(4)
                for j in range(2)]
    assert(fs.ls(ex) == expected)


@pytest.mark.skipif(os.environ.get('CI') == 'true', reason='Need GCS access')
def test_store_access(gcstemp, fs):
    paths = []
    with fs.store(gcstemp, ['ex1.txt', 'ex2.txt']) as datafiles:
        for d in datafiles:
            paths.append(d.path)
            d.handle.write(TEST_STRING)

    datafiles = fs.access(paths)
    for d in datafiles:
        assert(d.handle.read() == TEST_STRING)



import numpy as np
import blocks


def test_divide(datadir, randomdata, fs):
    blocks.divide(randomdata, datadir, 10, extension='.pkl')
    assert(len(fs.ls(datadir)) == 10)
    df = blocks.assemble(datadir)
    assert(np.isclose(df, randomdata).all().all())


def test_divide_offset(datadir, randomdata, fs):
    blocks.divide(randomdata, datadir, 10, extension='.pkl')
    blocks.divide(randomdata, datadir, 10, 10, extension='.pkl')
    assert(len(fs.ls(datadir)) == 20)
    df = blocks.assemble(datadir)
    expected = randomdata.append(randomdata)
    assert(np.isclose(df, expected).all().all())


def test_divide_cgroups(datadir, randomdata, fs):
    randomdata.insert(0, 'key', list(range(10)))
    cgroups_columns = {
        'cgroup1': ['key', 'f0', 'f1', 'f2'],
        'cgroup2': ['key', 'f3', 'f4', 'f5'],
        'cgroup3': ['key', 'f6', 'f7', 'f8', 'f9'],
    }

    blocks.divide(
        randomdata,
        datadir,
        10,
        cgroup_columns=cgroups_columns,
        extension='.pkl'
    )
    assert(len(fs.ls(datadir)) == 3)
    for subdir in fs.ls(datadir):
        assert(len(fs.ls(subdir)) == 10)
    df = blocks.assemble(datadir)
    assert(df.equals(randomdata))



import fnmatch
import os
import glob
import logging
import subprocess
import tempfile
import atexit
import shutil

from abc import ABCMeta, abstractmethod
from six import add_metaclass
from io import BytesIO
from collections import namedtuple
from contextlib import contextmanager
from google.cloud import storage

DataFile = namedtuple('DataFile', ['path', 'handle'])


@add_metaclass(ABCMeta)
class FileSystem(object):
    """ The required interface for any filesystem implementation

    See GCSFileSystem for a full implementation. This FileSystem is intended
    to be extendable to support cloud file systems, encryption strategies, etc...
    """
    @abstractmethod
    def ls(self, path):
        """ List files correspond to path, including glob wildcards

        Parameters
        ----------
        path : str
            The path to the file or directory to list; supports wildcards
        """
        pass

    @abstractmethod
    def access(self, paths):
        """ Access multiple paths as file-like objects

        This allows for optimization like parallel downloads

        Parameters
        ----------
        paths: list of str
            The paths of the files to access

        Returns
        -------
        files: list of DataFile
            A list of datafile instances, one for each input path
        """
        pass

    @abstractmethod
    def store(self, bucket, files):
        """ Store multiple data objects

        This allows for optimizations when storing several files

        Parameters
        ----------
        bucket : str
            The GCS bucket to use to store the files
        files : list of str
            The file names to store

        Returns
        -------
        datafiles : contextmanager
           A contextmanager that will yield datafiles and place them
           on the filesystem when finished
        """
        pass


class GCSFileSystem(FileSystem):
    """ File system interface that supports both local and GCS files

    This implementation uses subprocess and gsutil, which has excellent performance.
    However this can lead to problems in very multi-threaded applications and might not be
    as portable. For a python native implementation use GCSNativeFileSystem
    """
    GCS = 'gs://'

    def __init__(self, parallel=True, quiet=True):
        flags = []
        if parallel:
            flags.append('-m')
        if quiet:
            flags.append('-q')
        self.gcscp = ['gsutil'] + flags + ['cp', '-r']

    def local(self, path):
        """ Check if the path is available as a local file
        """
        return not path.startswith(self.GCS)

    def ls(self, path):
        """ List files correspond to path, including glob wildcards

        Parameters
        ----------
        path : str
            The path to the file or directory to list; supports wildcards
        """
        logging.info('Globbing file content in {}'.format(path))
        if not self.local(path):
            with open(os.devnull, 'w') as DEVNULL:
                p = subprocess.Popen(
                    ['gsutil', 'ls', path],
                    stdout=subprocess.PIPE,
                    stderr=DEVNULL,
                    universal_newlines=True,
                )
                stdout = p.communicate()[0]
            output = [line for line in stdout.split('\n') if line and line[-1] != ':']
        elif '**' in path:
            # Manual recursive glob, since in 2.X glob doesn't have recursive support
            path = path.rstrip('*')
            output = []
            for root, subdirs, files in os.walk(path):
                for fname in files:
                    output.append(os.path.join(root, fname))
        elif os.path.isdir(path):
            output = [os.path.join(path, f) for f in os.listdir(path)]
        else:
            output = glob.glob(path)
        return sorted(p.rstrip('/') for p in output)

    def copy(self, sources, dest):
        """ Copy the files in sources (recursively) to dest

        Parameters
        ----------
        sources : list of str
            The list of paths to copy, which can be directories
        dest : str
            The destination for the copy of source(s)
        """
        summary = ', '.join(sources)
        logging.info('Copying {} to {}...'.format(summary, dest))

        # TODO perhaps use a python API rather than subprocess
        if any(self.GCS in x for x in sources + [dest]):
            # at least one location is on GCS
            cmd = self.gcscp
        else:
            cmd = ['cp', '-r']
        # Break this into pieces so we don't hit OS limit for arguments
        # TODO: use getconf ARG_MAX and actually test number of bytes in sources
        CHUNK_SIZE = 1000
        sources_chunks = [sources[x:x+CHUNK_SIZE] for x in range(0, len(sources), CHUNK_SIZE)]
        for sources in sources_chunks:
            subprocess.check_call(cmd + sources + [dest])

    def access(self, paths):
        """ Access multiple paths as file-like objects

        This allows for optimization like parallel downloads

        Parameters
        ----------
        paths: list of str
            The paths of the files to access

        Returns
        -------
        files: list of DataFile
            A list of datafile instances, one for each input path
        """
        # Move the files into a tempdir from GCS
        tmpdir = _session_tempdir()
        self.copy(paths, tmpdir)

        # Then get file handles for each
        datafiles = []
        for path in paths:
            local = os.path.join(tmpdir, os.path.basename(path))
            datafiles.append(DataFile(path, open(local, 'rb')))
        return datafiles

    @contextmanager
    def store(self, bucket, files):
        """ Create file stores that will be written to the filesystem on close

        This allows for optimizations when storing several files

        Parameters
        ----------
        bucket : str
            The path of the bucket (on GCS) or folder (local) to store the data in
        files : list of str
            The filenames to create

        Returns
        -------
        datafiles : contextmanager
            A context manager that yields datafiles and when the context is closed
            they are written to GCS

        Usage
        -----
        >>> with filesystem.store('gs://bucket/sub/', ['ex1.txt', 'ex2.txt']) as datafiles:
        >>>     datafiles[0].handle.write('example 1')
        >>>     datafiles[1].handle.write('example 2')
        """
        # Make local files in a tempdir that serve as the file handles
        tmpdir = _session_tempdir()
        datafiles = []
        local_files = []
        for f in files:
            local = os.path.join(tmpdir, f)
            local_files.append(local)
            datafiles.append(DataFile(os.path.join(bucket, f), open(local, 'wb')))

        yield datafiles

        for d in datafiles:
            d.handle.close()

        if self.local(bucket) and not os.path.exists(bucket):
            os.makedirs(bucket)

        self.copy(local_files, os.path.join(bucket, ''))


class GCSNativeFileSystem(GCSFileSystem):
    """ File system interface that supports GCS and local files

    This uses the native python cloud storage library for read and write, rather than gsutil.
    The performance is significantly slower when reading/writing several files but is thread-safe
    for applications which are already parallelized. It also stores the files entirely in
    memory rather than using tempfiles.
    """
    def __init__(self, *args, **kwargs):
        self.client = storage.Client()
        super(GCSNativeFileSystem, self).__init__(*args, **kwargs)

    def ls(self, path):
        logging.info('Globbing file content in {}'.format(path))

        # use GCSFileSystem's implementation for local paths
        if self.local(path):
            return super(GCSNativeFileSystem, self).ls(path)

        # find all names that start with the prefix for this path
        bucket, path = self._split(path)
        prefix = self._prefix(path)
        names = [b.name for b in self.client.get_bucket(bucket).list_blobs(prefix=prefix)]

        if prefix == path:
            # no pattern strings, listing a single file or a folder
            try:
                # one blob exactly matches the requested path
                names = [next(p for p in names if p == path)]
            except StopIteration:
                # no exact match, so we are listing based on prefix
                # find files in this directory or subfolders
                names = self._list_single(prefix, names)
        # we have a pattern to match
        else:
            # for recursive glob, do not attempt to match folders, only files
            if '**' in path:
                names = fnmatch.filter(names, path)
            else:
                # make sure we can match folders in addition to blobs
                candidates = set(names) | set(self._list_single(prefix, names))
                names = fnmatch.filter(candidates, path)
                # one more filter because fnmatch recurses single *
                names = [name for name in names if name.count('/') == path.count('/')]

        paths = ['gs://{}/{}'.format(bucket, name) for name in names]
        return sorted(paths)

    def _prefix(self, path):
        # find the lowest prefix that doesn't include a pattern match
        splits = path.split('/')
        accumulate = []
        while splits:
            sub = splits.pop(0)
            if any(x in sub for x in ['*', '?', '[', ']']):
                break
            accumulate.append(sub)
        return '/'.join(accumulate) if accumulate else None

    def _list_single(self, prefix, names):
        # find files that are directly in the dir specified by prefix, not in subfolders
        valid = set(n.replace(prefix, '').lstrip('/').split('/')[0] for n in names)
        return [os.path.join(prefix, n) for n in valid]

    def access(self, paths):
        """ Access multiple paths as file-like objects

        This allows for optimization like parallel downloads

        Parameters
        ----------
        paths: list of str
            The paths of the files to access

        Returns
        -------
        files: list of DataFile
            A list of datafile instances, one for each input path
        """
        datafiles = []
        for path in paths:
            datafiles.append(DataFile(path, BytesIO()))
        for datafile in datafiles:
            self._read(datafile)
        return datafiles

    @contextmanager
    def store(self, bucket, files):
        """ Create file stores that will be written to the filesystem on close

        This allows for optimizations when storing several files

        Parameters
        ----------
        bucket : str
            The path of the bucket (on GCS) or folder (local) to store the data in
        files : list of str
            The filenames to create

        Returns
        -------
        datafiles : contextmanager
            A context manager that yields datafiles and when the context is closed
            they are written to GCS

        Usage
        -----
        >>> with filesystem.store('gs://bucket/sub/', ['ex1.txt', 'ex2.txt']) as datafiles:
        >>>     datafiles[0].handle.write('example 1')
        >>>     datafiles[1].handle.write('example 2')
        """
        # Make BytesIO instances that serve as the file handles
        datafiles = []
        for f in files:
            datafiles.append(DataFile(os.path.join(bucket, f), BytesIO()))

        yield datafiles

        for d in datafiles:
            d.handle.seek(0)
            self._write(d)

    def _split(self, path):
        bucket = path.replace(self.GCS, '').split('/')[0]
        prefix = "gs://{}".format(bucket)
        path = path[len(prefix) + 1:]
        return bucket, path

    def _blob(self, path):
        bucket, path = self._split(path)
        return storage.Blob(path, self.client.get_bucket(bucket))

    def _read(self, datafile):
        if self.local(datafile.path):
            with open(datafile.path, 'rb') as f:
                datafile.handle.write(f.read())
        else:
            self._blob(datafile.path).download_to_file(datafile.handle)
        datafile.handle.seek(0)

    def _write(self, datafile):
        if self.local(datafile.path):
            with open(datafile.path, 'w') as f:
                f.write(datafile.handle.read())
        else:
            self._blob(datafile.path).upload_from_file(datafile.handle)


def _session_tempdir():
    """ Create a tempdir that will be cleaned up at session exit
    """
    tmpdir = tempfile.mkdtemp()
    # create and use a subdir of specified name to preserve cgroup logic
    atexit.register(lambda: shutil.rmtree(tmpdir))
    return tmpdir



import json
import os
import pandas as pd
import numpy as np
import pytest
import six
import blocks.dfio as io
from blocks.filesystem import DataFile


def test_read_write_native_formats(randomdata, datadir_local):
    extensions = ['.pkl', '.csv', '.pkl.gz', '.csv.gz']
    for extension in extensions:
        path = os.path.join(datadir_local, 'tmp{}'.format(extension))
        df = _reload(randomdata, path)
        assert(np.isclose(df, randomdata).all().all())


def test_compression_csv(randomdata, datadir_local):
    path = os.path.join(datadir_local, 'tmp.csv.gz')
    # write compressed with pandas
    randomdata.to_csv(path, index=False, compression='gzip')

    # read compressed with blocks
    with open(path, 'rb') as f:
        d = DataFile(path, f)
        df = io.read_df(d)
    assert(np.isclose(df, randomdata).all().all())

    # write compressed with blocks
    with open(path, 'wb') as f:
        d = DataFile(path, f)
        io.write_df(randomdata, d)
    # read compressed with pandas
    df = pd.read_csv(path, compression='gzip')
    assert(np.isclose(df, randomdata).all().all())


def test_compression_pickle(randomdata, datadir_local):
    path = os.path.join(datadir_local, 'tmp.pkl.gz')
    # write compressed with pandas
    randomdata.to_pickle(path, compression='gzip')

    # read compressed with blocks
    with open(path, 'rb') as f:
        d = DataFile(path, f)
        df = io.read_df(d)
    assert(np.isclose(df, randomdata).all().all())

    # write compressed with blocks
    with open(path, 'wb') as f:
        d = DataFile(path, f)
        io.write_df(randomdata, d)
    # read compressed with pandas
    df = pd.read_pickle(path, compression='gzip')
    assert(np.isclose(df, randomdata).all().all())


def test_compression_parquet(randomdata, datadir_local):
    pytest.importorskip('pyarrow')
    pytest.importorskip('pandas', minversion='0.22.0')
    path = os.path.join(datadir_local, 'tmp.parquet.gz')
    # write compressed with pandas
    randomdata.to_parquet(path, compression='gzip')

    # read compressed with blocks
    with open(path, 'rb') as f:
        d = DataFile(path, f)
        df = io.read_df(d)
    assert(np.isclose(df, randomdata).all().all())

    # write compressed with blocks
    with open(path, 'wb') as f:
        d = DataFile(path, f)
        io.write_df(randomdata, d)
    # read compressed with pandas
    df = pd.read_parquet(path)
    assert(np.isclose(df, randomdata).all().all())


def test_read_write_parquet(randomdata, datadir_local):
    pytest.importorskip('pyarrow')
    pytest.importorskip('pandas', minversion='0.22.0')
    path = os.path.join(datadir_local, 'tmp.pq')
    df = _reload(randomdata, path)
    assert(np.isclose(df, randomdata).all().all())


def test_read_write_parquet_unicode(randomdata, datadir_local):
    pytest.importorskip('pyarrow')
    pytest.importorskip('pandas', minversion='0.22.0')
    randomdata[u'f10'] = randomdata['f9']
    path = os.path.join(datadir_local, 'tmp{}'.format('.pq'))
    if six.PY2:
        # expect a warning about this in python2
        with pytest.warns(UserWarning):
            df = _reload(randomdata, path)
    else:
        df = _reload(randomdata, path)
    assert(np.isclose(df, randomdata).all().all())


def test_read_write_avro(randomdata, datadir_local):
    pytest.importorskip('fastavro')
    path = os.path.join(datadir_local, 'tmp.avro')
    schema_path = os.path.join(datadir_local, 'tmp.avsc')
    _write_schema(randomdata, schema_path)
    df = _reload(randomdata, path, schema_path=schema_path)
    assert(np.isclose(df, randomdata).all().all())


def _write_schema(data, path):
    schema = {"type": "record", "name": "testschema"}
    schema["fields"] = [{"name": str(c), "type": "float"} for c in data.columns.values]
    with open(path, 'w') as f:
        json.dump(schema, f)


def _reload(data, path, **kwargs):
    with open(path, 'wb') as f:
        d = DataFile(path, f)
        io.write_df(data, d, **kwargs)
    with open(path, 'rb') as f:
        d = DataFile(path, f)
        df = io.read_df(d)
    return df



import os
import numpy as np
import blocks


def test_place(datadir, randomdata, fs):
    blocks.place(randomdata, os.path.join(datadir, 'example.csv'))
    assert(len(fs.ls(datadir)) == 1)
    df = blocks.assemble(datadir)
    assert(np.isclose(df, randomdata).all().all())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from blocks import core


def test_assemble(populated):
    df = core.assemble(populated)
    assert(df.shape == (40, 41))
    expected = ['f{}_{}'.format(i, j) for i in range(4) for j in range(10)]
    expected.append('key')
    assert(set(df.columns) == set(expected))


def test_assemble_ordered(populated):
    order = ['c2', 'c1', 'c3', 'c0']
    df = core.assemble(populated, cgroups=order)
    assert(df.shape == (40, 41))
    expected = ['f{}_{}'.format(order[i][1], j) for i in range(4) for j in range(10)]
    expected.append('key')
    assert(set(df.columns) == set(expected))
    # Check the features are in the right order
    assert([c for c in df.columns if c != 'key'] == expected[:-1])


# Various options do not depend on filesystem so we can just test locally
def test_assemble_filtered_cgroup(populated_local, keys):
    df = core.assemble(populated_local, cgroups=['c0', 'c3'])
    assert(df.shape == (40, 21))
    expected = ['f{}_{}'.format(i, j) for i in [0, 3] for j in range(10)]
    expected.append('key')
    assert(set(df.columns) == set(expected))
    assert((df.key == keys).all())


def test_assemble_filtered_rgroup(populated_local, keys):
    df = core.assemble(populated_local, rgroups=['part.0.csv', 'part.1.csv'])
    assert(df.shape == (20, 41))
    expected = ['f{}_{}'.format(i, j) for i in range(4) for j in range(10)]
    expected.append('key')
    assert(set(df.columns) == set(expected))
    assert((df.key == keys[:20].reset_index(drop=True)).all())


def test_assemble_read_args(populated_local, keys):
    read_args = {'dtype': str}
    df = core.assemble(populated_local, read_args=read_args)
    assert(df.shape == (40, 41))
    expected = ['f{}_{}'.format(i, j) for i in range(4) for j in range(10)]
    expected.append('key')
    assert(set(df.columns) == set(expected))
    assert((df.key == keys).all())
    assert((df.dtypes == 'object').all())


def test_assemble_cgroup_args(populated_local, keys):
    cgroup_args = {'c0': {'dtype': str}}
    df = core.assemble(populated_local, cgroup_args=cgroup_args)
    assert(df.shape == (40, 41))
    expected = ['f{}_{}'.format(i, j) for i in range(4) for j in range(10)]
    expected.append('key')
    assert(set(df.columns) == set(expected))
    assert((df.key == keys).all())
    for col in ['f0_{}'.format(i) for i in range(10)]:
        assert(df.dtypes[col] == 'object')



# -*- coding: utf-8 -*-

__author__ = """Bradley Axen"""
__email__ = 'baxen@squareup.com'
__version__ = '0.5.2'


from blocks.core import assemble, iterate, partitioned, divide, place  # NOQA
from blocks.dfio import read_df, write_df  # NOQA



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
from blocks import core


def test_expand(populated, fs):
    expected = [os.path.join(populated, 'c{}/part.{}.csv'.format(i, j))
                for i in range(4)
                for j in range(4)]

    # All of these patterns should expand into the same set of files
    for ex in ['', '*', '*/*', '**']:
        ex = os.path.join(populated, ex)
        paths = fs.ls(ex)
        expanded = sorted(core._expand(paths, fs))
        assert(expanded == expected)


def test_expand_pattern(populated, fs):
    expected = [os.path.join(populated, 'c{}/part.{}.csv'.format(i, j))
                for i in range(2)
                for j in range(4)]

    # All of these patterns should expand into the same set of files
    for ex in ['c[01]', 'c[01]/*']:
        ex = os.path.join(populated, ex)
        paths = fs.ls(ex)
        expanded = sorted(core._expand(paths, fs))
        assert(expanded == expected)


def test_cgroups():
    expanded = [
        'base/c{}/part.{}.csv'.format(i, j)
        for i in range(4)
        for j in range(4)
    ]
    cgroups = core._cgroups(expanded)
    for i in range(4):
        key = 'c{}'.format(i)
        assert(key in cgroups)
        assert(cgroups[key] == ['base/c{}/part.{}.csv'.format(i, j) for j in range(4)])


def test_access(populated, fs):
    paths = fs.ls(populated)
    expanded = core._expand(paths, fs)
    cgroups = core._cgroups(expanded)
    cgroups = core._access(cgroups, fs)
    assert(len(cgroups) == 4)
    for c, datafiles in cgroups.items():
        assert(len(datafiles) == 4)
        for d in datafiles:
            assert(d.handle.read())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

with open('README.md') as readme_file:
    readme = readme_file.read()

requirements = [
    'six',
    'pandas',
    'google-cloud-storage',
]

extras_require = {
    'pq': ['pyarrow'],
    'avro': ['fastavro'],
    'tests': ['pytest', 'pytest-cov', 'delegator.py', 'flake8'],
    'doc': ['sphinx', 'numpydoc', 'sphinx-rtd-theme']
}
extras_require['dev'] = set(sum(extras_require.values(), []))

setup(
    name='sq-blocks',
    version='0.5.2',
    description=(
        'Blocks provides a simple interface to read, organize, and manipulate structured data'
        ' in files on local and cloud storage'
    ),
    long_description=readme,
    long_description_content_type="text/markdown",
    author='Bradley Axen',
    author_email='baxen@squareup.com',
    packages=find_packages(),
    include_package_data=True,
    install_requires=requirements,
    extras_require=extras_require,
    zip_safe=False,
    keywords='blocks',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
    ],
)
